
?-ensure_loaded(converter_nfa_to_dfa).


get_distinct_states(Distinct,[],_,Distinct):-!.
get_distinct_states(TempDistinct,LastAdded,Delta,Distinct):-
   findall([X1,X2],Symbol^(member([Y1,Y2],LastAdded),    /* maybe to slow */
                    member([X2,Symbol,Y2],Delta),
                    member([X1,Symbol,Y1],Delta),
                    not(member([X1,X2],TempDistinct))),
                 NewDistinct),
   append(NewDistinct,TempDistinct,NextDistinct),
   get_distinct_states(NextDistinct,NewDistinct,Delta,Distinct).



get_distinct_states(dfa(Q,_,Delta,_,F),Distinct):-
   mergesort(Q,QS),
   mergesort(F,FS),
   set_subtraction_sorted_sets(QS,FS,NonFS),
   findall([X,Y],((member(X,FS),member(Y,NonFS));(member(Y,FS),member(X,NonFS))),ElementaryDistinct),
   get_distinct_states(ElementaryDistinct,ElementaryDistinct,Delta,Distinct).



collect_partitions(QTemp, [State1|Q1Rest], Distinct, QResult):-
   findall(EquivalentState,(member(EquivalentState,Q1Rest),
                           not(member([EquivalentState,State1],Distinct))),
                        NewPartition),
   !,
   set_subtraction_sorted_sets(Q1Rest,NewPartition,NextQ),
   collect_partitions([[State1|NewPartition]|QTemp],NextQ,Distinct,QResult).
collect_partitions(Q,_,_,Q).



optimize_dfa_sets(dfa(Q1,Sigma,Delta1,S1,F1), dfa(Q,Sigma,Delta,S,F)):-
   get_distinct_states(dfa(Q1,Sigma,Delta1,S1,F1),Distinct),
   collect_partitions([],Q1,Distinct,Q), /* Q is a set of sets */
   findall([Start,Symbol,Target],(member(Start,Q), /* pick start partition */
                                  member(Target,Q), /* pick target partition */
                                  member(StartState,Start), /* which contain states */
                                  member(TargetState,Target),
                                  member([StartState,Symbol,TargetState],Delta1)),
                                                          /* that are connected in Delta */
                                 DeltaDoubles),
   remove_doubles(DeltaDoubles,Delta),
   /* the partition that contains S1 */
   member(S,Q),
   member(S1,S),
   /* the partitions where the first element is in F1 (i.e. all elements of them
   are in F1, since they are equivalent, because they are in the same partition) */
   findall([First|Sets],(member([First|Sets],Q),member(First,F1)),F).



get_reachable_states(Q,Delta,Epsilon,LastAdded,TempReachable, Reachable):-
   bagof(NextLevel,Symbol^(  member(Start,LastAdded),
                      (  member([Start,Symbol,NextLevel],Delta);
                         member([Start,NextLevel],Epsilon)),
                      not(member(NextLevel,TempReachable)) ), NewReachables ),
   !,
   append(NewReachables,TempReachable,NextReachable),
   get_reachable_states(Q,Delta,Epsilon,NewReachables,NextReachable,Reachable).
get_reachable_states(_,_,_,_,Reachable, Reachable).



remove_unreachable_states(epsilon_nfa(Q1,Sigma,Delta1,Epsilon1,S,F1),epsilon_nfa(Q,Sigma,Delta,Epsilon,S,F)):-
   mergesort(Q1,QS),
   mergesort(F1,FS),
   get_reachable_states(QS,Delta1,Epsilon1,[S],[S], Reachable),
   make_sorted_set(Reachable,ReachableSet),
   set_intersection_sorted_sets(QS,ReachableSet,Q),
   set_intersection_sorted_sets(FS,ReachableSet,F),
   findall([Start,Symbol,Target], (  member([Start,Symbol,Target],Delta1),member(Start,Q),member(Target,Q)  ), Delta),
   findall([Start,Target], (  member([Start,Target],Epsilon1),member(Start,Q),member(Target,Q)  ), Epsilon).



optimize_dfa(dfa(Q1,Sigma1,Delta1,S1,F1), DFA):-
   /* no nondeterminism is added */
   remove_unreachable_states(epsilon_nfa(Q1,Sigma1,Delta1,[],S1,F1),epsilon_nfa(Q2,Sigma2,Delta2,_,S2,F2)),
   optimize_dfa_sets(dfa(Q2,Sigma2,Delta2,S2,F2),SetDFA),
   set_dfa_to_dfa(SetDFA,DFA).

